                     /*-----------------------------------*/
                     /*             >>>Slip<<<            */
                     /*            Theo D'Hondt           */
                     /*     VUB Software Languages Lab    */
                     /*             (c) 2010              */
                     /*-----------------------------------*/
                     /*   version 2: using a trampoline   */
                     /*-----------------------------------*/
                     /*              evaluate             */
                     /*-----------------------------------*/

#include <stdio.h>

#include "SlipMain.h"
#include "SlipDictionary.h"
#include "SlipEvaluate.h"

/*------------------------------ private macros ------------------------------*/

#define chunk_size(TYP)                                                        \
  ((sizeof(TYP) - sizeof(CNT))/ sizeof(CEL_type))

/*------------------------------ private constants ---------------------------*/

static const TXT_type AL1_error_string = "at least one expression required";
static const TXT_type BRQ_error_string = "boolean value required";
static const TXT_type E1X_error_string = "exactly one expression required";
static const TXT_type ILT_error_string = "illegal expression type";
static const TXT_type ITA_error_string = "improperly terminated application";
static const TXT_type ITF_error_string = "improperly terminated form";
static const TXT_type ITP_error_string = "improperly terminated procedure";
static const TXT_type IPA_error_string = "invalid parameter";
static const TXT_type IVP_error_string = "invalid pattern in form";
static const TXT_type IVV_error_string = "missing variable in form";
static const TXT_type IXT_error_string = "invalid expression type";
static const TXT_type MSP_error_string = "missing pattern in form";
static const TXT_type MSV_error_string = "invalid variable in form";
static const TXT_type PNR_error_string = "procedure or native required";
static const TXT_type TFO_error_string = "too few operands in application";
static const TXT_type TFX_error_string = "too few expressions in form";
static const TXT_type TMO_error_string = "too many operands in application";
static const TXT_type TMX_error_string = "too many expressions in form";

const static UNS_type Exit_counter_preset = 256;

/*------------------------------ private prototypes --------------------------*/

static NIL_type        evaluate_bindings(PRC_type,
                                         PAI_type,
                                         CNT_type,
                                         ENV_type);
static NIL_type            evaluate_body(PRC_type,
                                         CNT_type,
                                         ENV_type,
                                         ENV_type);
static NIL_type      evaluate_expression(EXP_type,
                                         CNT_type,
                                         ENV_type);
static NIL_type     evaluate_native_call(NAT_type,
                                         PAI_type,
                                         CNT_type,
                                         ENV_type);
static NIL_type        evaluate_sequence(PAI_type,
                                         CNT_type,
                                         ENV_type);
static NIL_type evaluate_vararg_bindings(PRC_type,
                                         SYM_type,
                                         PAI_type,
                                         ENV_type,
                                         CNT_type,
                                         ENV_type);

/*------------------------------ private variables ---------------------------*/

static UNS_type Exit_counter;

/*------------------------------ private functions ---------------------------*/

static NIL_type continue_with(CNT_type Continuation,
                              EXP_type Value,
                              ENV_type Environment)
  { CCC_type function;
    CFN_type c_function;
    c_function = Continuation->cfn;
    function = c_function->ccc;
    function(Value,
             Continuation,
             Environment); }

static SGN_type list_length(PAI_type List)
  { PAI_type pair;
    SGN_type count;
    count = 0;
    for (pair = List;
         is_PAI(pair);
         pair = pair->cdr)
      count++;
    if (is_NUL(pair))
      return count;
    return -1; }

/*----------------------------------------------------------------------------*/
/*--------------------------------- application ------------------------------*/
/*----------------------------------------------------------------------------*/

static const TXT_type Application_String = "application";

static CFN_type Continue_application;

typedef struct aPL * aPL_type;
typedef struct aPL { CEL_type hdr;
                     CFN_type cfn;
                     CNT_type cnt;
                     PAI_type opd; } aPL;

static const UNS_type aPL_size = chunk_size(aPL);

static NIL_type continue_application(EXP_type Procedure,
                                     CNT_type Continuation,
                                     ENV_type Environment)
  { aPL_type application_thread;
    CNT_type continuation;
    PAI_type operands;
    TAG_type tag;
    application_thread = (aPL_type)Continuation;
    continuation = application_thread->cnt;
    operands     = application_thread->opd;
    tag = Tag_of(Procedure);
    switch (tag)
      { case NAT_tag:
          evaluate_native_call(Procedure,
                               operands,
                               continuation,
                               Environment); 
        case PRC_tag:
             evaluate_bindings(Procedure,
                               operands,
                               continuation,
                               Environment); }
    Main_Error_Text(PNR_error_string,
                    Application_String); }

/*----------------------------------------------------------------------------*/

static NIL_type evaluate_application(EXP_type Operator,
                                     EXP_type Operands,
                                     CNT_type Continuation,
                                     ENV_type Environment)
  { aPL_type application_thread;
    CNT_type continuation;
    continuation = make_CNT(Continue_application,
                            Continuation,
                            aPL_size);
    application_thread = (aPL_type)continuation;
    application_thread->opd = Operands;
    evaluate_expression(Operator,
                        continuation,
                        Environment); }

static NIL_type initialize_application(NIL_type)
  { Continue_application = make_CFN(continue_application); }
    
/*----------------------------------------------------------------------------*/
/*------------------------------------ begin ---------------------------------*/
/*----------------------------------------------------------------------------*/

static NIL_type evaluate_begin(PAI_type Operands,
                               CNT_type Continuation,
                               ENV_type Environment)
  { TAG_type tag;
    tag = Tag_of(Operands);
    switch (tag)
      { case NUL_tag:
          Main_Error_Text(AL1_error_string,
                          Main_Begin_String); 
        case PAI_tag:
          evaluate_sequence(Operands,
                            Continuation,
                            Environment); }
    Main_Error_Text(ITF_error_string,
                    Main_Begin_String); }

/*----------------------------------------------------------------------------*/
/*---------------------------------- bindings --------------------------------*/
/*----------------------------------------------------------------------------*/

static CFN_type Continue_bindings;

typedef struct bND * bND_type;
typedef struct bND { CEL_type hdr;
                     CFN_type cfn;
                     CNT_type cnt;
                     PRC_type prc;
                     PAI_type par;
                     PAI_type opd;
                     ENV_type clo; } bND;

static const UNS_type bND_size = chunk_size(bND);

static NIL_type binding(PRC_type Procedure,
                        EXP_type Parameters,
                        PAI_type Operands,
                        ENV_type Closure,
                        CNT_type Continuation,
                        ENV_type Environment)
  { bND_type binding_thread;
    CNT_type continuation;
    ENV_type closure;
    EXP_type expression;
    PAI_type operands;
    TAG_type tag;
    tag = Tag_of(Parameters);
    switch (tag)
      { case NUL_tag:
          if (is_NUL(Operands))
            evaluate_body(Procedure,
                          Continuation,
                          Environment,
                          Closure); 
          Main_Error_Procedure(TMO_error_string,
                               Procedure);
        case PAI_tag:
          tag = Tag_of(Operands);
          switch (tag)
            { case NUL_tag:
                Main_Error_Procedure(TFO_error_string,
                                     Procedure);
              case PAI_tag:
                expression = Operands->car;
                operands   = Operands->cdr;
                continuation = make_CNT(Continue_bindings,
                                        Continuation,
                                        bND_size);
                binding_thread = (bND_type)continuation;
                binding_thread->prc = Procedure;
                binding_thread->par = Parameters;
                binding_thread->opd = operands;
                binding_thread->clo = Closure;
                evaluate_expression(expression,
                                    continuation,
                                    Environment); }
          Main_Error_Procedure(ITA_error_string,
                               Procedure);
        case SYM_tag:
          tag = Tag_of(Operands);
          switch (tag)
            { case NUL_tag:
                closure = Dictionary_Define(Parameters,
                                            Main_Empty_Environment,
                                            Closure);
                evaluate_body(Procedure,
                              Continuation,
                              Environment,
                              closure);
              case PAI_tag:
                evaluate_vararg_bindings(Procedure,
                                         Parameters,
                                         Operands,
                                         Closure,
                                         Continuation,
                                         Environment); }
          Main_Error_Procedure(ITA_error_string,
                               Procedure); }
    Main_Error_Procedure(ITP_error_string,
                         Procedure); }

static NIL_type continue_bindings(EXP_type Value,
                                  CNT_type Continuation,
                                  ENV_type Environment)
  { bND_type binding_thread;
    CNT_type continuation;
    ENV_type closure;
    EXP_type parameter;
    PAI_type operands,
             parameters;
    PRC_type procedure;
    binding_thread = (bND_type)Continuation;
    continuation = binding_thread->cnt;
    procedure    = binding_thread->prc;
    parameters   = binding_thread->par;
    operands     = binding_thread->opd;
    closure      = binding_thread->clo;
    parameter = parameters->car;
    if (is_SYM(parameter))
      { parameters = parameters->cdr;
        closure = Dictionary_Define(parameter,
                                    Value,
                                    closure);
        binding(procedure,
                parameters,
                operands,
                closure,
                continuation,
                Environment); }
    Main_Error_Procedure(IPA_error_string,
                         procedure); }

/*----------------------------------------------------------------------------*/

static NIL_type evaluate_bindings(PRC_type Procedure,
                                  PAI_type Operands,
                                  CNT_type Continuation,
                                  ENV_type Environment)
  { ENV_type closure;
    EXP_type parameters;
    parameters = Procedure->par;
    closure    = Procedure->clo;
    binding(Procedure,
            parameters,
            Operands,
            closure,
            Continuation,
            Environment); }
                
static NIL_type initialize_bindings(NIL_type)
  { Continue_bindings = make_CFN(continue_bindings); }
    
/*----------------------------------------------------------------------------*/
/*-------------------------------------- body --------------------------------*/
/*----------------------------------------------------------------------------*/

static CFN_type Continue_body;

typedef struct bOD * bOD_type;
typedef struct bOD { CEL_type hdr;
                     CFN_type cfn;
                     CNT_type cnt;
                     ENV_type env; } bOD;

static const UNS_type bOD_size = chunk_size(bOD);

static NIL_type continue_body(EXP_type Value,
                              CNT_type Continuation,
                              ENV_type Ignore)
  { bOD_type body_thread;
    CNT_type continuation;
    ENV_type environment;
    body_thread = (bOD_type)Continuation;
    continuation = body_thread->cnt;
    environment  = body_thread->env;
    continue_with(continuation,
                  Value,
                  environment); }

/*----------------------------------------------------------------------------*/

static NIL_type evaluate_body(PRC_type Procedure,
                              CNT_type Continuation,
                              ENV_type Environment,
                              ENV_type Closure)
  { bOD_type body_thread;
    CNT_type continuation;
    PAI_type body;
    continuation = make_CNT(Continue_body,
                            Continuation,
                            bOD_size);
    body_thread = (bOD_type)continuation;
    body_thread->env = Environment;
    body = Procedure->bod;
    evaluate_sequence(body,
                      continuation,
                      Closure); }

static NIL_type evaluate_inline(EXP_type Expression,
                                CNT_type Continuation,
                                ENV_type Environment)
  { bOD_type body_thread;
    CNT_type continuation;
    continuation = make_CNT(Continue_body,
                            Continuation,
                            bOD_size);
    body_thread = (bOD_type)continuation;
    body_thread->env = Environment;
    evaluate_expression(Expression,
                        continuation,
                        Environment); }

static NIL_type evaluate_inline_sequence(PAI_type Expressions,
                                         CNT_type Continuation,
                                         ENV_type Environment)
  { bOD_type body_thread;
    CNT_type continuation;
    continuation = make_CNT(Continue_body,
                            Continuation,
                            bOD_size);
    body_thread = (bOD_type)continuation;
    body_thread->env = Environment;
    evaluate_sequence(Expressions,
                      continuation,
                      Environment); }

static NIL_type initialize_body(NIL_type)
  { Continue_body = make_CFN(continue_body); }

/*----------------------------------------------------------------------------*/
/*------------------------------------ define --------------------------------*/
/*----------------------------------------------------------------------------*/

static CFN_type Continue_define;

typedef struct dEF * dEF_type;
typedef struct dEF { CEL_type hdr;
                     CFN_type cfn;
                     CNT_type cnt;
                     SYM_type var; } dEF;

static const UNS_type dEF_size = chunk_size(dEF);

static NIL_type continue_define(EXP_type Value,
                                CNT_type Continuation,
                                ENV_type Environment)
  { dEF_type define_thread;
    CNT_type continuation;
    ENV_type environment;
    SYM_type variable;
    define_thread = (dEF_type)Continuation;
    continuation = define_thread->cnt;
    variable     = define_thread->var;
    environment = Dictionary_Define(variable,
                                    Value,
                                    Environment);
    continue_with(continuation,
                  Value,
                  environment); }

/*----------------------------------------------------------------------------*/

static NIL_type evaluate_define(PAI_type Operands,
                                CNT_type Continuation,
                                ENV_type Environment)
  { dEF_type define_thread;
    CNT_type continuation;
    ENV_type environment;
    EXP_type expression,
             pattern;
    PAI_type body,
             expressions,
             parameters,
             residue;
    PRC_type procedure;
    SYM_type variable;
    TAG_type tag;
    tag = Tag_of(Operands);
    switch (tag)
      { case NUL_tag:
          Main_Error_Text(MSP_error_string,
                          Main_Define_String);
        case PAI_tag:
          pattern = Operands->car;
          body    = Operands->cdr;
          tag = Tag_of(body);
          switch (tag)
            { case NUL_tag:
                Main_Error_Text(AL1_error_string,
                                Main_Define_String);
              case PAI_tag:
                tag = Tag_of(pattern);
                switch (tag)
                  { case PAI_tag:
                      expressions = (PAI_type)pattern;
                      variable = expressions->car;
                      if (is_SYM(variable))
                        { parameters = expressions->cdr;
                          procedure = make_PRC(variable,
                                               parameters,
                                               body,
                                               Main_Empty_Environment);
                          environment = Dictionary_Define(variable,
                                                          procedure,
                                                          Environment);
                          procedure->clo = environment;
                          continue_with(Continuation,
                                        procedure,
                                        environment); }
                      Main_Error_Text(IVV_error_string,
                                      Main_Define_String);
                    case SYM_tag:
                      variable = (SYM_type)pattern;
                      expression = body->car;
                      residue    = body->cdr;
                      if (is_NUL(residue))
                        { continuation = make_CNT(Continue_define,
                                                  Continuation,
                                                  dEF_size);
                          define_thread = (dEF_type)continuation;
                          define_thread->var = variable;
                          evaluate_expression(expression,
                                              continuation,
                                              Environment); }
                      Main_Error_Text(E1X_error_string,
                                      Main_Define_String); }
                Main_Error_Text(IVP_error_string,
                                Main_Define_String); }
          Main_Error_Text(ITF_error_string,
                          Main_Define_String); }
    Main_Error_Text(ITF_error_string,
                    Main_Define_String); }

static NIL_type initialize_define(NIL_type)
  { Continue_define = make_CFN(continue_define); }

/*----------------------------------------------------------------------------*/
/*------------------------------- direct_application -------------------------*/
/*----------------------------------------------------------------------------*/

static const TXT_type Direct_Application_String = "direct application";

static NIL_type evaluate_direct_application(EXP_type Value,
                                            PAI_type Arguments,
                                            CNT_type Continuation,
                                            ENV_type Environment)
  { EXP_type pattern,
             value;
    ENV_type closure;
    FUN_type function;
    NAT_type native;
    PAI_type arguments,
             parameters;
    PRC_type procedure;
    SGN_type length;
    SYM_type variable;
    TAG_type tag;
    UNS_type index;
    VEC_type vector;
    tag = Tag_of(Value);
    switch (tag)
      { case NAT_tag:
          native = (NAT_type)Value;
          function = native->fun;
          length = list_length(Arguments);
          switch (length)
            { case 0:
                vector = Main_Empty_Vector;
                break;
              case -1:
                { TXT_type name = native->nam;
                  Main_Error_Text(ITA_error_string,
                                  name); }
              default:
                { arguments = Arguments;
                  vector = make_VEC(length);
                  for (index = 1;
                       index <= length;
                       index++)
                    { value = arguments->car;
                      arguments = arguments->cdr;
                      vector[index] = value; }
                  break; }}
          function(vector,
                   Continuation,
                   Environment); 
        case PRC_tag:
          arguments = Arguments;
          procedure = (PRC_type)Value;
          pattern     = procedure->par;
          closure     = procedure->clo;
          for (;;)
            { tag = Tag_of(pattern);
              switch (tag)
                { case NUL_tag:
                    if (is_NUL(arguments))
                      evaluate_body(procedure,
                                    Continuation,
                                    Environment,
                                    closure);
                    Main_Error_Procedure(TMO_error_string,
                                         procedure);
                  case PAI_tag:
                    tag = Tag_of(arguments);
                    switch (tag)
                      { case NUL_tag:
                          Main_Error_Procedure(TFO_error_string,
                                               procedure);
                        case PAI_tag:
                          value = arguments->car;
                          parameters = pattern;
                          variable = parameters->car;
                          if (is_SYM(variable))
                            { arguments = arguments->cdr;
                              pattern   = parameters->cdr;
                              closure = Dictionary_Define(variable,
                                                          value,
                                                          closure);
                              continue; }
                          Main_Error_Procedure(IPA_error_string,
                                               procedure); }
                    Main_Error_Procedure(ITA_error_string,
                                         procedure);
                  case SYM_tag:
                    variable = pattern;
                    closure = Dictionary_Define(variable,
                                                arguments,
                                                closure);
                    evaluate_body(procedure,
                                  Continuation,
                                  Environment,
                                  closure); }
              Main_Error_Procedure(ITP_error_string,
                                   procedure); }}
    Main_Error_Text(PNR_error_string,
                    Direct_Application_String); }
                    
/*----------------------------------------------------------------------------*/
/*------------------------------------ if ------------------------------------*/
/*----------------------------------------------------------------------------*/

static CFN_type Continue_if_double;

typedef struct iFD * iFD_type;
typedef struct iFD { CEL_type hdr;
                     CFN_type cfn;
                     CNT_type cnt;
                     EXP_type cns;
                     EXP_type alt; } iFD;

static const UNS_type iFD_size = chunk_size(iFD);

static NIL_type continue_if_double(EXP_type Boolean,
                                   CNT_type Continuation,
                                   ENV_type Environment)
  { iFD_type if_double_thread;
    CNT_type continuation;
    EXP_type alternative,
             consequent;
    TAG_type tag;
    tag = Tag_of(Boolean);
    if_double_thread = (iFD_type)Continuation;
    continuation = if_double_thread->cnt;
    switch (tag)
      { case FLS_tag:
          alternative = if_double_thread->alt;
          evaluate_inline(alternative,
                          continuation,
                          Environment);
        case TRU_tag:
          consequent = if_double_thread->cns;
          evaluate_inline(consequent,
                          continuation,
                          Environment); }
    Main_Error_Text(BRQ_error_string,
                    Main_If_String); }

/*----------------------------------------------------------------------------*/

static CFN_type Continue_if_single;

typedef struct iFS * iFS_type;
typedef struct iFS { CEL_type hdr;
                     CFN_type cfn;
                     CNT_type cnt;
                     EXP_type cns; } iFS;

static const UNS_type iFS_size = chunk_size(iFS);

static NIL_type continue_if_single(EXP_type Boolean,
                                   CNT_type Continuation,
                                   ENV_type Environment)
  { iFS_type if_single_thread;
    CNT_type continuation;
    EXP_type consequent;
    TAG_type tag;
    tag = Tag_of(Boolean);
    if_single_thread = (iFS_type)Continuation;
    continuation = if_single_thread->cnt;
    switch (tag)
      { case FLS_tag:
          continue_with(continuation,
                        Main_Unspecified,
                        Environment);
        case TRU_tag:
          consequent = if_single_thread->cns;
          evaluate_inline(consequent,
                          continuation,
                          Environment); }
    Main_Error_Text(BRQ_error_string,
                    Main_If_String); }

/*----------------------------------------------------------------------------*/

static NIL_type evaluate_if(PAI_type Operands,
                            CNT_type Continuation,
                            ENV_type Environment)
  { iFD_type if_double_thread;
    iFS_type if_single_thread;
    CNT_type continuation;
    EXP_type alternative,
             consequent,
             predicate,
             residue;
    PAI_type alternatives,
             clauses;
    TAG_type tag;
    tag = Tag_of(Operands);
    switch (tag)
      { case NUL_tag:
          Main_Error_Text(TFX_error_string,
                          Main_If_String);
        case PAI_tag:
          predicate = Operands->car;
          clauses   = Operands->cdr;
          tag = Tag_of(clauses);
          switch (tag)
            { case NUL_tag:
                Main_Error_Text(TFX_error_string,
                                Main_If_String);
              case PAI_tag:
                consequent   = clauses->car;
                alternatives = clauses->cdr;
                tag = Tag_of(alternatives);
                switch (tag)
                  { case NUL_tag:
                      continuation = make_CNT(Continue_if_single,
                                                            Continuation,
                                                            iFS_size);
                      if_single_thread = (iFS_type)continuation;
                      if_single_thread->cns = consequent;
                      evaluate_expression(predicate,
                                          continuation,
                                          Environment);
                    case PAI_tag:
                      alternative = alternatives->car;
                      residue     = alternatives->cdr;
                      if (is_NUL(residue))
                        { continuation = make_CNT(Continue_if_double,
                                                  Continuation,
                                                  iFD_size);
                          if_double_thread = (iFD_type)continuation;
                          if_double_thread->cns = consequent;
                          if_double_thread->alt = alternative;
                          evaluate_expression(predicate,
                                              continuation,
                                              Environment); }
                      Main_Error_Text(TMX_error_string,
                                      Main_If_String); }
                Main_Error_Text(ITF_error_string,
                                Main_If_String); }
          Main_Error_Text(ITF_error_string,
                          Main_If_String); }
    Main_Error_Text(ITF_error_string,
                    Main_If_String); }

static NIL_type initialize_if(NIL_type)
  { Continue_if_double = make_CFN(continue_if_double);
    Continue_if_single = make_CFN(continue_if_single); }

/*----------------------------------------------------------------------------*/
/*------------------------------------ lambda --------------------------------*/
/*----------------------------------------------------------------------------*/

static NIL_type evaluate_lambda(PAI_type Operands,
                                CNT_type Continuation,
                                ENV_type Environment)
  { EXP_type parameters;
    PAI_type body;
    PRC_type procedure;
    TAG_type tag;
    tag = Tag_of(Operands);
    switch (tag)
      { case NUL_tag:
          Main_Error_Text(TFX_error_string,
                          Main_Lambda_String);
        case PAI_tag:
          parameters = Operands->car;
          body       = Operands->cdr;
          tag = Tag_of(body);
          switch (tag)
            { case NUL_tag:
                Main_Error_Text(AL1_error_string,
                                Main_Lambda_String);
              case PAI_tag:
                procedure = make_PRC(Main_Anonymous,
                                     parameters,
                                     body,
                                     Environment);
                continue_with(Continuation,
                              procedure,
                              Environment); }
          Main_Error_Text(ITF_error_string,
                          Main_Lambda_String); }
    Main_Error_Text(ITF_error_string,
                    Main_Lambda_String); }

/*----------------------------------------------------------------------------*/
/*-------------------------------- native call -------------------------------*/
/*----------------------------------------------------------------------------*/

static CFN_type Continue_native_call;

typedef struct nCL * nCL_type;
typedef struct nCL { CEL_type hdr;
                     CFN_type cfn;
                     CNT_type cnt;
                     PAI_type exs;
                     VEC_type vec;
                     NAT_type nat;
                     NBR_type pos; } nCL;

static const UNS_type nCL_size = chunk_size(nCL);

static NIL_type native_call(PAI_type Expressions,
                            VEC_type Vector,
                            NAT_type Native,
                            UNS_type Index,
                            CNT_type Continuation,
                            ENV_type Environment)
  { nCL_type native_call_thread;
    CNT_type continuation;
    EXP_type expression;
    NBR_type position;
    PAI_type expressions;
    expression  = Expressions->car;
    expressions = Expressions->cdr;
    position = make_NBR(Index);
    continuation = make_CNT(Continue_native_call,
                            Continuation,
                            nCL_size);
    native_call_thread = (nCL_type)continuation;
    native_call_thread->exs = expressions;
    native_call_thread->vec = Vector;
    native_call_thread->nat = Native;
    native_call_thread->pos = position;
    evaluate_expression(expression,
                        continuation,
                        Environment); }

static NIL_type continue_native_call(EXP_type Value,
                                     CNT_type Continuation,
                                     ENV_type Environment)
  { nCL_type native_call_thread;
    CNT_type continuation;
    FUN_type function;
    NAT_type native;
    NBR_type position;
    PAI_type expressions;
    UNS_type index,
             length;
    VEC_type vector;
    native_call_thread = (nCL_type)Continuation;
    continuation = native_call_thread->cnt;
    vector       = native_call_thread->vec;
    native       = native_call_thread->nat;
    position     = native_call_thread->pos;
    index = position->lng;
    length = size_VEC(vector);
    vector[index] = Value;
    if (index < length)
      { expressions = native_call_thread->exs;
        native_call(expressions,
                    vector,
                    native,
                    index + 1,
                    continuation,
                    Environment); }
    function = native->fun;
    function(vector,
             continuation,
             Environment); }

/*----------------------------------------------------------------------------*/

static NIL_type evaluate_native_call(NAT_type Native,
                                     PAI_type Operands,
                                     CNT_type Continuation,
                                     ENV_type Environment)
  { FUN_type function;
    SGN_type length;
    TXT_type name;
    VEC_type vector;
    length = list_length(Operands);
    switch (length)
      { case 0:
          function = Native->fun;
          function(Main_Empty_Vector,
                   Continuation,
                   Environment);
        case -1:
          name = Native->nam;
          Main_Error_Text(ITA_error_string,
                          name); }
    vector = make_VEC(length);
    native_call(Operands,
                vector,
                Native,
                1,
                Continuation,
                Environment); }

static NIL_type initialize_native_call(NIL_type)
  { Continue_native_call = make_CFN(continue_native_call); }
  
/*----------------------------------------------------------------------------*/
/*------------------------------------ quote ---------------------------------*/
/*----------------------------------------------------------------------------*/

static NIL_type evaluate_quote(PAI_type Operands,
                               CNT_type Continuation,
                               ENV_type Environment)
  { EXP_type expression,
             residue;
    expression = Operands->car;
    residue    = Operands->cdr;
    if (is_NUL(residue))
      continue_with(Continuation,            /* clone !!! */
                    expression,
                    Environment);
    Main_Error_Text(TMX_error_string,
                    Main_Quote_String); }

/*----------------------------------------------------------------------------*/
/*---------------------------------- sequence --------------------------------*/
/*----------------------------------------------------------------------------*/

static CFN_type Continue_sequence;

typedef struct sEQ * sEQ_type;
typedef struct sEQ { CEL_type hdr;
                     CFN_type cfn;
                     CNT_type cnt;
                     PAI_type exs; } sEQ;

static const UNS_type sEQ_size = chunk_size(sEQ);

static NIL_type continue_sequence(EXP_type Value,
                                  CNT_type Continuation,
                                  ENV_type Environment)
  { sEQ_type sequence_thread;
    CNT_type continuation;
    PAI_type expressions;
    sequence_thread = (sEQ_type)Continuation;
    continuation = sequence_thread->cnt;
    expressions  = sequence_thread->exs;
    evaluate_sequence(expressions,
                      continuation,
                      Environment); }

/*----------------------------------------------------------------------------*/

static NIL_type evaluate_sequence(PAI_type Expressions,
                                  CNT_type Continuation,
                                  ENV_type Environment)
  { sEQ_type sequence_thread;
    CNT_type continuation;
    EXP_type expression;
    PAI_type expressions;
    TAG_type tag;
    expression  = Expressions->car;
    expressions = Expressions->cdr;
    tag = Tag_of(expressions);
    switch (tag)
      { case NUL_tag:
          evaluate_expression(expression,
                              Continuation,
                              Environment);
        case PAI_tag:
          continuation = make_CNT(Continue_sequence,
                                  Continuation,
                                  sEQ_size);
          sequence_thread = (sEQ_type)continuation;
          sequence_thread->exs = expressions;
          evaluate_expression(expression,
                              continuation,
                              Environment); }
    Main_Error_Text(ITF_error_string,
                    Main_Sequence_String); }

static NIL_type initialize_sequence(NIL_type)
  { Continue_sequence = make_CFN(continue_sequence); }

/*----------------------------------------------------------------------------*/
/*------------------------------------ set -----------------------------------*/
/*----------------------------------------------------------------------------*/

static CFN_type Continue_set;

typedef struct sET * sET_type;
typedef struct sET { CEL_type hdr;
                     CFN_type cfn;
                     CNT_type cnt;
                     SYM_type var; } sET;

static const UNS_type sET_size = chunk_size(sET);

static NIL_type continue_set(EXP_type Value,
                             CNT_type Continuation,
                             ENV_type Environment)
  { sET_type set_thread;
    CNT_type continuation;
    SYM_type variable;
    set_thread = (sET_type)Continuation;
    continuation = set_thread->cnt;
    variable     = set_thread->var;
    Dictionary_Replace(variable,
                       Value,
                       Environment);
    continue_with(continuation,
                  Value,
                  Environment); }

/*----------------------------------------------------------------------------*/

static NIL_type evaluate_set(PAI_type Operands,
                             CNT_type Continuation,
                             ENV_type Environment)
  { sET_type set_thread;
    CNT_type continuation;
    EXP_type expression;
    PAI_type expressions,
             residue;
    SYM_type variable;
    TAG_type tag;
    tag = Tag_of(Operands);
    switch (tag)
      { case NUL_tag:
          Main_Error_Text(MSV_error_string,
                          Main_Set_String);
        case PAI_tag:
          variable = Operands->car;
          if (is_SYM(variable))
            { expressions = Operands->cdr;
              tag = Tag_of(expressions);
              switch (tag)
                { case NUL_tag:
                    Main_Error_Text(E1X_error_string,
                                    Main_Set_String);
                  case PAI_tag:
                    expression = expressions->car;
                    residue    = expressions->cdr;
                    if (is_NUL(residue))
                      { continuation = make_CNT(Continue_set,
                                                Continuation,
                                                sET_size);
                        set_thread = (sET_type)continuation;
                        set_thread->var = variable;
                        evaluate_expression(expression,
                                            continuation,
                                            Environment); }
                    Main_Error_Text(TMX_error_string,
                                    Main_Set_String); }
              Main_Error_Text(ITF_error_string,
                              Main_Set_String); }
          Main_Error_Text(IVV_error_string,
                          Main_Set_String); }
    Main_Error_Text(ITF_error_string,
                    Main_Set_String); }

static NIL_type initialize_set(NIL_type)
  { Continue_set = make_CFN(continue_set); }

/*----------------------------------------------------------------------------*/
/*------------------------------------ symbol --------------------------------*/
/*----------------------------------------------------------------------------*/

static NIL_type evaluate_symbol(SYM_type Variable,
                                CNT_type Continuation,
                                ENV_type Environment)
  { EXP_type value;
    value = Dictionary_Lookup(Variable,
                              Environment);
    continue_with(Continuation,
                  value,
                  Environment); }

/*----------------------------------------------------------------------------*/
/*----------------------------------- value ----------------------------------*/
/*----------------------------------------------------------------------------*/

static NIL_type evaluate_value(EXP_type Value,
                               CNT_type Continuation,
                               ENV_type Environment)
  { continue_with(Continuation,
                  Value,
                  Environment); }

/*----------------------------------------------------------------------------*/
/*------------------------------- vararg_bindings ----------------------------*/
/*----------------------------------------------------------------------------*/

static CFN_type Continue_vararg_bindings;

typedef struct vRG * vRG_type;
typedef struct vRG { CEL_type hdr;
                     CFN_type cfn;
                     CNT_type cnt;
                     PRC_type prc;
                     SYM_type par;
                     PAI_type opd;
                     PAI_type arg;
                     ENV_type clo; } vRG;

static const UNS_type vRG_size = chunk_size(vRG);

static NIL_type vararg_bindings(PRC_type Procedure,
                                SYM_type Parameter,
                                PAI_type Operands,
                                EXP_type Arguments,
                                ENV_type Closure,
                                CNT_type Continuation,
                                ENV_type Environment)
  { vRG_type vararg_bindings_thread;
    CNT_type continuation;
    ENV_type closure;
    EXP_type expression;
    PAI_type arguments,
             operands;
    TAG_type tag;
    tag = Tag_of(Operands);
    switch (tag)
      { case NUL_tag:
          arguments = Main_Reverse(Arguments);
          closure = Dictionary_Define(Parameter,
                                      arguments,
                                      Closure);
          evaluate_body(Procedure,
                        Continuation,
                        Environment,
                        closure);
        case PAI_tag:
          expression = Operands->car;
          operands   = Operands->cdr;
          continuation = make_CNT(Continue_vararg_bindings,
                                  Continuation,
                                  vRG_size);
          vararg_bindings_thread = (vRG_type)continuation;
          vararg_bindings_thread->prc = Procedure;
          vararg_bindings_thread->par = Parameter;
          vararg_bindings_thread->opd = operands;
          vararg_bindings_thread->arg = Arguments;
          vararg_bindings_thread->clo = Closure;
          evaluate_expression(expression,
                              continuation,
                              Environment); }
    Main_Error_Procedure(ITA_error_string,
                         Procedure); }

static NIL_type continue_vararg_bindings(EXP_type Value,
                                         CNT_type Continuation,
                                         ENV_type Environment)
  { vRG_type vararg_bindings_thread;
    CNT_type continuation;
    ENV_type closure;
    PAI_type arguments,
             operands;
    PRC_type procedure;
    SYM_type parameter;
    vararg_bindings_thread = (vRG_type)Continuation;
    continuation = vararg_bindings_thread->cnt;
    procedure    = vararg_bindings_thread->prc;
    parameter    = vararg_bindings_thread->par;
    operands     = vararg_bindings_thread->opd;
    arguments    = vararg_bindings_thread->arg;
    closure      = vararg_bindings_thread->clo;
    arguments = make_PAI(Value,
                         arguments);
    vararg_bindings(procedure,
                    parameter,
                    operands,
                    arguments,
                    closure,
                    continuation,
                    Environment); }

/*----------------------------------------------------------------------------*/

static NIL_type evaluate_vararg_bindings(PRC_type Procedure,
                                         SYM_type Parameter,
                                         PAI_type Operands,
                                         ENV_type Closure,
                                         CNT_type Continuation,
                                         ENV_type Environment)
  { vararg_bindings(Procedure,
                    Parameter,
                    Operands,
                    Main_Empty_Pair,
                    Closure,
                    Continuation,
                    Environment); }

static NIL_type initialize_vararg_bindings(NIL_type)
  { Continue_vararg_bindings = make_CFN(continue_vararg_bindings); }

/*----------------------------------------------------------------------------*/
/*----------------------------------- while ----------------------------------*/
/*----------------------------------------------------------------------------*/

static CFN_type Continue_while_body;

typedef struct wHB * wHB_type;
typedef struct wHB { CEL_type hdr;
                     CFN_type cfn;
                     CNT_type cnt;
                     EXP_type prd;
                     PAI_type bod;
                     ENV_type env; } wHB;

static const UNS_type wHB_size = chunk_size(wHB);

static NIL_type while_predicate(CNT_type,
                                EXP_type,
                                PAI_type,
                                ENV_type,
                                EXP_type);

static NIL_type continue_while_body(EXP_type Result,
                                    CNT_type Continuation,
                                    ENV_type Environment)
  { wHB_type while_body_thread;
    CNT_type continuation;
    ENV_type environment;
    EXP_type body,
             predicate;
    while_body_thread = (wHB_type)Continuation;
    continuation = while_body_thread->cnt;
    predicate    = while_body_thread->prd;
    body         = while_body_thread->bod;
    environment  = while_body_thread->env;
    while_predicate(continuation,
                    predicate,
                    body,
                    environment,
                    Result); }

/*----------------------------------------------------------------------------*/

static CFN_type Continue_while_predicate;

typedef struct wHP * wHP_type;
typedef struct wHP { CEL_type hdr;
                     CFN_type cfn;
                     CNT_type cnt;
                     EXP_type prd;
                     PAI_type bod;
                     ENV_type env;
                     EXP_type res; } wHP;

static const UNS_type wHP_size = chunk_size(wHP);

static NIL_type while_predicate(CNT_type Continuation,
                                EXP_type Predicate,
                                PAI_type Body,
                                ENV_type Environment,
                                EXP_type Result)
  { wHP_type while_predicate_thread;
    CNT_type continuation;
    continuation = make_CNT(Continue_while_predicate,
                            Continuation,
                            wHP_size);
    while_predicate_thread = (wHP_type)continuation;
    while_predicate_thread->prd = Predicate;
    while_predicate_thread->bod = Body;
    while_predicate_thread->env = Environment;
    while_predicate_thread->res = Result;
    evaluate_inline(Predicate,
                    continuation,
                    Environment); }

static NIL_type continue_while_predicate(EXP_type Boolean,
                                         CNT_type Continuation,
                                         ENV_type Environment)
  { wHB_type while_body_thread;
    wHP_type while_predicate_thread;
    CNT_type continuation;
    ENV_type environment;
    EXP_type predicate,
             result;
    PAI_type body;
    while_predicate_thread = (wHP_type)Continuation;
    continuation = while_predicate_thread->cnt;
    environment  = while_predicate_thread->env;
    if (is_FLS(Boolean))
      { result = while_predicate_thread->res;
        continue_with(continuation,
                      result,
                      environment); }
    predicate = while_predicate_thread->prd;
    body      = while_predicate_thread->bod;
    continuation = make_CNT(Continue_while_body,
                            continuation,
                            wHB_size);
    while_body_thread = (wHB_type)continuation;
    while_body_thread->prd = predicate;
    while_body_thread->bod = body;
    while_body_thread->env = environment;
    evaluate_inline_sequence(body,
                             continuation,
                             environment); }

/*----------------------------------------------------------------------------*/

static NIL_type evaluate_while(PAI_type Operands,
                               CNT_type Continuation,
                               ENV_type Environment)
  { EXP_type body,
             predicate;
    TAG_type tag;
    tag = Tag_of(Operands);
    switch (tag)
      { case NUL_tag:
          Main_Error_Text(TFX_error_string,
                          Main_While_String);
        case PAI_tag:
          predicate = Operands->car;
          body = Operands->cdr;
          tag = Tag_of(body);
          switch (tag)
            { case NUL_tag:
                Main_Error_Text(TFX_error_string,
                                Main_While_String);
              case PAI_tag:
                while_predicate(Continuation,
                                predicate,
                                body,
                                Environment,
                                Main_Unspecified); }
          Main_Error_Text(ITF_error_string,
                          Main_While_String); }
    Main_Error_Text(ITF_error_string,
                    Main_While_String); }

static NIL_type initialize_while(NIL_type)
  { Continue_while_predicate = make_CFN(continue_while_predicate);
    Continue_while_body = make_CFN(continue_while_body); }

/*----------------------------------------------------------------------------*/
/*------------------------------------ form ----------------------------------*/
/*----------------------------------------------------------------------------*/

static NIL_type evaluate_form(PAI_type Form,
                              CNT_type Continuation,
                              ENV_type Environment)
  { EXP_type operands,
             operator;
    operator = Form->car;
    operands = Form->cdr;
    if (operator == Main_Begin)
          evaluate_begin(operands,
                         Continuation,
                         Environment);
    if (operator == Main_Define)
         evaluate_define(operands,
                         Continuation,
                         Environment);
    if (operator == Main_If)
             evaluate_if(operands,
                         Continuation,
                         Environment);
    if (operator == Main_Lambda)
         evaluate_lambda(operands,
                         Continuation,
                         Environment);
    if (operator == Main_Quote)
          evaluate_quote(operands,
                         Continuation,
                         Environment);
    if (operator == Main_Set)
            evaluate_set(operands,
                         Continuation,
                         Environment);
    if (operator == Main_While)
          evaluate_while(operands,
                         Continuation,
                         Environment);
    evaluate_application(operator,
                         operands,
                         Continuation,
                         Environment); }

/*----------------------------------------------------------------------------*/
/*------------------------------------ expression ----------------------------*/
/*----------------------------------------------------------------------------*/

static NIL_type evaluate_expression(EXP_type Expression,
                                    CNT_type Continuation,
                                    ENV_type Environment)
  { TAG_type tag;
    if (--Exit_counter == 0)
      { Exit_counter = Exit_counter_preset;
        Main_Exit(Expression,
                  Continuation,
                  Environment); }
    tag = Tag_of(Expression);
    switch (tag)
      { case PAI_tag:
            evaluate_form(Expression,
                          Continuation,
                          Environment);
        case SYM_tag:
          evaluate_symbol(Expression,
                          Continuation,
                          Environment);
        case CHA_tag:
        case FLS_tag:
        case NUL_tag:
        case NBR_tag:
        case REA_tag:
        case STR_tag:
        case TRU_tag:
        case VEC_tag:
           evaluate_value(Expression,
                          Continuation,
                          Environment); 
        case CFN_tag:
        case CNT_tag:
        case ENV_tag:
        case NAT_tag:
        case PRC_tag:
        case USP_tag:
          Main_Error_Tag(ILT_error_string,
                         tag); }
    Main_Error_Tag(IXT_error_string,
                   tag); }

/*------------------------------ public functions ----------------------------*/

NIL_type Evaluate_Initialize(NIL_type)
  { initialize_application();
    initialize_bindings();
    initialize_body();
    initialize_define();
    initialize_if();
    initialize_native_call();
    initialize_sequence();
    initialize_set();
    initialize_vararg_bindings();
    initialize_while(); }

NIL_type Evaluate_Continue(EXP_type Value,
                           CNT_type Continuation,
                           ENV_type Environment)
  { continue_with(Continuation,
                  Value,
                  Environment); }

NIL_type Evaluate_Apply(EXP_type Procedure,
                        PAI_type Arguments,
                        CNT_type Continuation,
                        ENV_type Environment)
  { evaluate_direct_application(Procedure,
                                Arguments,
                                Continuation,
                                Environment); }

NIL_type Evaluate_Evaluate(EXP_type Expression,
                           CNT_type Continuation,
                           ENV_type Environment)
  { Exit_counter = Exit_counter_preset;
    evaluate_expression(Expression,
                        Continuation,
                        Environment); }
