                     /*-----------------------------------*/
                     /*             >>>Slip<<<            */
                     /*            Theo D'Hondt           */
                     /*     VUB Software Languages Lab    */
                     /*             (c) 2010              */
                     /*-----------------------------------*/
                     /*  version 1: straightforward code  */
                     /*-----------------------------------*/
                     /*               grammar             */
                     /*-----------------------------------*/

#include "SlipMemory.h"

/*------------------------------ public types --------------------------------*/

typedef enum { CFN_tag ,                                      /* C-function   */
               CHA_tag ,                                      /* character    */
               CNT_tag ,                                      /* continuation */
               ENV_tag ,                                      /* environment  */
               FLS_tag ,                                      /* false        */
               NAT_tag ,                                      /* native       */
               NBR_tag ,                                      /* number       */
               NUL_tag ,                                      /* null         */
               PAI_tag ,                                      /* pair         */
               PRC_tag ,                                      /* procedure    */
               REA_tag ,                                      /* real         */
               STR_tag ,                                      /* string       */
               SYM_tag ,                                      /* symbol       */
               TRU_tag ,                                      /* true         */
               USP_tag ,                                      /* unspecified  */
               VEC_tag } TAG_type;                            /* vector       */

typedef struct CFN * CFN_type;
typedef struct CHA * CHA_type;
typedef struct CNT * CNT_type;
typedef struct ENV * ENV_type;
typedef struct FLS * FLS_type;
typedef struct NAT * NAT_type;
typedef struct NBR * NBR_type;
typedef struct NUL * NUL_type;
typedef struct PAI * PAI_type;
typedef struct PRC * PRC_type;
typedef struct REA * REA_type;
typedef struct STR * STR_type;
typedef struct SYM * SYM_type;
typedef struct TRU * TRU_type;
typedef struct USP * USP_type;

typedef   NIL_type * EXP_type;

typedef   EXP_type * VEC_type;

typedef NIL_type (* CCC_type) (EXP_type,
                               CNT_type,
                               ENV_type);
typedef NIL_type (* FUN_type) (VEC_type,
                               CNT_type,
                               ENV_type);

/*-------------------------------- public functions --------------------------*/

TAG_type Tag_of(EXP_type);

/*------------------------------ public declarations -------------------------*/

typedef
  struct CFN { CEL_type hdr;
               CCC_type ccc; } CFN;
BYT_type   is_CFN(EXP_type);
CFN_type make_CFN(CCC_type);

typedef
  struct CHA { CEL_type hdr;
               CHR_type chr; } CHA;
BYT_type   is_CHA(EXP_type);
CHA_type make_CHA(CHR_type);

typedef
  struct CNT { CEL_type hdr;
               CFN_type cfn;
               CNT_type cnt;
               EXP_type exp[]; } CNT;
BYT_type   is_CNT(EXP_type);
CNT_type make_CNT(CFN_type,
                  CNT_type,
                  UNS_type);

typedef
  struct ENV { CEL_type hdr;
               SYM_type var;
               EXP_type val;
               ENV_type env; } ENV;
BYT_type   is_ENV(EXP_type);
ENV_type make_ENV(SYM_type,
                  EXP_type,
                  ENV_type);

typedef
  struct FLS { CEL_type hdr; } FLS;
BYT_type   is_FLS(EXP_type);
FLS_type make_FLS(NIL_type);

typedef
  struct NAT { CEL_type hdr;
               FUN_type fun;
               CHR_type nam[]; } NAT;
BYT_type   is_NAT(EXP_type);
NAT_type make_NAT(FUN_type,
                  TXT_type);

typedef
  struct NBR { CEL_type hdr;
               LNG_type lng; } NBR;
BYT_type   is_NBR(EXP_type);
NBR_type make_NBR(LNG_type);

typedef
  struct NUL { CEL_type hdr; } NUL;
BYT_type   is_NUL(EXP_type);
NUL_type make_NUL(NIL_type);

typedef
  struct PAI { CEL_type hdr;
               EXP_type car;
               EXP_type cdr; } PAI;
BYT_type   is_PAI(EXP_type);
PAI_type make_PAI(EXP_type,
                  EXP_type);

typedef
  struct PRC { CEL_type hdr;
               SYM_type nam;
               EXP_type par;
               EXP_type bod;
               ENV_type clo; } PRC;
BYT_type   is_PRC(EXP_type);
PRC_type make_PRC(SYM_type,
                  EXP_type,
                  EXP_type,
                  ENV_type);

typedef
  struct REA { CEL_type hdr;
               FLO_type flo; } REA;
BYT_type   is_REA(EXP_type);
REA_type make_REA(FLO_type);

typedef
  struct STR { CEL_type hdr;
               CHR_type txt[]; } STR;
BYT_type   is_STR(EXP_type);
STR_type make_STR(TXT_type);

typedef
  struct SYM { CEL_type hdr;
               CHR_type txt[]; } SYM;
BYT_type   is_SYM(EXP_type);
SYM_type make_SYM(TXT_type);

typedef
  struct TRU { CEL_type hdr; } TRU;
BYT_type   is_TRU(EXP_type);
TRU_type make_TRU(NIL_type);

typedef
  struct USP { CEL_type hdr; } USP;
BYT_type   is_USP(EXP_type);
USP_type make_USP(NIL_type);

BYT_type   is_VEC(EXP_type);
VEC_type make_VEC(UNS_type);
UNS_type size_VEC(VEC_type);
